import json
import logging
import tokenize
from abc import abstractmethod, ABC
from io import StringIO
from pathlib import Path
from typing import Optional

import pyarrow as pa
import os

import numpy as np
import pandas as pd
from omf import OMFReader, VolumeGridGeometry
import pyarrow.parquet as pq
from pandera import DataFrameSchema


class BaseReader(ABC):

    def __init__(self, file_path: Path):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.file_path: Path = file_path
        self.variables_in_file: list[str] = []
        self.records_in_file: int = 0

    @staticmethod
    def _parse_query_columns(query) -> list[str]:
        # Create a list to store the column names
        column_names = []

        # Tokenize the query string
        for token in tokenize.generate_tokens(StringIO(query).readline):
            token_type, token_string, _, _, _ = token

            # If the token is a name, and it's not a built-in Python name, add it to the list
            if token_type == tokenize.NAME and token_string not in __builtins__:
                column_names.append(token_string)

        return column_names

    @abstractmethod
    def read(self, columns: Optional[list[str]] = None, query: Optional[str] = None) -> pd.DataFrame:
        pass

    @abstractmethod
    def get_index(self) -> pd.Index:
        pass

    def validate(self, schema_file: Path, data: Optional[pd.DataFrame]) -> pd.DataFrame:
        """Validate using a pandera schema

        This method does not leverage multiprocessing, and loads the entire dataframe into memory.
        Args:
            schema_file: The path to the schema yaml file
            data: The data to validate, if not provided, the underlying read method will be called.
        Returns:
            The coerced DataFrame after validation
        """
        import pandera as pa
        schema: DataFrameSchema = pa.DataFrameSchema.from_yaml(schema_file)
        if data:
            df = data
        else:
            df = self.read()
        schema.validate(df, lazy=True, inplace=True)
        return df

    def preprocess(self, negative_to_nan_threshold: Optional[float] = -1,
                   not_detected_assays_threshold: Optional[float] = 0.5) -> pd.DataFrame:
        """
        Preprocess the data by managing negative values.
        Args:
            negative_to_nan_threshold: Values below this threshold will be replaced with NaN
            not_detected_assays_threshold: Values above this threshold will be replaced with half the absolute value

        Returns:
            The preprocessed DataFrame, with no negatives and no values above the not_detected_assays_threshold.

        """
        if negative_to_nan_threshold > 0:
            raise ValueError("The negative_to_nan_threshold must be less than or equal to zero or None.")
        if not_detected_assays_threshold > 0:
            raise ValueError("The not_detected_assays_threshold must be less than or equal to zero or None")

        df = self.read()

        # detect numeric columns
        numeric_cols = df.select_dtypes(include=[np.number]).columns

        if negative_to_nan_threshold:
            df.loc[df[numeric_cols] < negative_to_nan_threshold, numeric_cols] = np.nan
        if not_detected_assays_threshold:
            mask = (df[numeric_cols] > not_detected_assays_threshold) and (df[numeric_cols] < 0)
            df.loc[mask, numeric_cols] = np.abs(df.loc[mask, numeric_cols]) / 2
        return df


class ParquetFileReader(BaseReader):
    """
    Read a Parquet file
    """

    def __init__(self, file_path: Path):
        """
        Initialize the parquet reader.  While not enforced, it is expected that the file is indexed by x, y, z, or
        x, y, z, dx, dy, dz
        Args:
            file_path: The path to the Parquet file.
        """
        super().__init__(file_path)
        self.variables_in_file = self._get_parquet_columns()
        self.records_in_file = self._get_parquet_length()

    def _get_parquet_columns(self):
        parquet_file = pq.ParquetFile(self.file_path)
        metadata: dict = self.get_parquet_metadata()
        cols = [col for col in parquet_file.schema.names if col not in metadata['index_columns']]
        return cols

    def _get_parquet_length(self):
        parquet_file = pq.ParquetFile(self.file_path)
        return parquet_file.metadata.num_rows

    def get_parquet_metadata(self) -> dict:
        parquet_file = pq.ParquetFile(self.file_path)
        pd_metadata_bytes = parquet_file.metadata.metadata.get(b'pandas')
        pd_metadata_str: str = pd_metadata_bytes.decode('utf-8')
        return json.loads(pd_metadata_str)

    def get_index(self) -> pd.Index:
        parquet_file = pq.ParquetFile(self.file_path)
        pd_metadata: dict = self.get_parquet_metadata()
        index_columns = pd_metadata['index_columns']
        # deal with the single range index case
        if len(index_columns) == 1:
            if index_columns[0].get('kind') == 'range':
                df_index = pd.Index(
                    range(index_columns[0].get('start'), index_columns[0].get('stop'), index_columns[0].get('step')))
            else:
                df_index = pd.Index(parquet_file.read(columns=index_columns[0].get('name')).to_pandas())
        else:
            # extract the pd.MultiIndex
            df_index = parquet_file.read(columns=index_columns).to_pandas().index
        return df_index

    def read(self, columns: Optional[list[str]] = None, query: Optional[str] = None,
             with_index: bool = True) -> pd.DataFrame:
        # If no columns are specified, load all columns
        if not columns:
            columns = self.variables_in_file
        else:
            # Check if the columns specified are valid
            for col in columns:
                if col not in self.variables_in_file:
                    raise ValueError(f"Column '{col}' not found in the Parquet file: {self.file_path}.  "
                                     f"Available columns are: {self.variables_in_file}")

        # If a query is specified, parse it to find the columns involved
        if query:
            query_columns = self._parse_query_columns(query)
            # Load only the columns involved in the query
            parquet_file = pq.ParquetFile(self.file_path)
            df_query = parquet_file.read(columns=query_columns).to_pandas()  # Apply the query to the DataFrame
            df_query = df_query.query(query)
            # Get the indices of the rows that match the query
            query_indices = df_query.index
            # Load the remaining columns, but only for the rows that match the query
            remaining_columns = [col for col in columns if col not in query_columns]
            if remaining_columns:
                chunks = []
                for col in remaining_columns:
                    df_col = parquet_file.read(columns=[col]).to_pandas()
                    chunks.append(df_col.loc[query_indices])
                # Concatenate the query DataFrame and the remaining DataFrame
                df = pd.concat([df_query, *chunks], axis=1)
            else:
                df = df_query
            if with_index:
                df_index: pd.Index = self.get_index()[query_indices]
                df.set_index(df_index, inplace=True, drop=True)

        else:
            # If no query is specified, load the specified columns
            df = pd.read_parquet(self.file_path, columns=columns)
            if with_index is False:
                df.reset_index(drop=True, inplace=True)

        return df


class OMFFileReader(BaseReader):
    """
    Read an OMF file
    """

    def __init__(self, file_path, element: str):
        """
        Initialize the OMF file reader.  The element must be a VolumeElement in the OMF file.
        Args:
            file_path: The path to the OMF file
            element: The name of the element in the OMF file to be validated. E.g. 'Block Model'
        """
        super().__init__(file_path)

        # check that the element provided is a valid VolumeElement in the OMF file.
        self.elements = OMFReader(str(file_path)).get_project().elements
        self.element_names = [e.name for e in self.elements]
        if element not in self.element_names:
            raise ValueError(f"Element '{element}' not found in the OMF file: {file_path}. Available elements are:"
                             f" {list(self.elements.keys())}")
        elif self.get_element_by_name(element).__class__.__name__ != 'VolumeElement':
            raise ValueError(f"Element '{element}' is not a VolumeElement in the OMF file: {file_path}")

        self.element = self.get_element_by_name(element)

        self.variables_in_file = [v.name for v in self.element.data]
        self.records_in_file = len(self.element.data[0].array.array)

    def get_element_by_name(self, element_name: str):
        # get the index of the element in order to index into elements
        element_index = self.element_names.index(element_name)
        return self.elements[element_index]

    def read(self, columns: Optional[list[str]] = None, query: Optional[str] = None,
             with_index: bool = True) -> pd.DataFrame:
        # Get the VolumeElement from the OMF file
        # volume_element = OMFReader(self.file_path).get_project().elements[self.element]

        # If no columns are specified, load all columns
        if not columns:
            columns = self.variables_in_file
        else:
            # Check if the columns specified are valid
            for col in columns:
                if col not in self.variables_in_file:
                    raise ValueError(f"Column '{col}' not found in the VolumeElement: {self.element}")

        # If a query is specified, parse it to find the columns involved
        if query:
            query_columns = self._parse_query_columns(query)
            # Load only the columns involved in the query
            df_query: pd.DataFrame = self.read_volume_variables(self.element, variables=query_columns)
            # Apply the query to the DataFrame
            df_query = df_query.query(query)
            # Get the indices of the rows that match the query
            query_indices = df_query.index
            # Load the remaining columns, but only for the rows that match the query
            remaining_columns = [col for col in columns if col not in query_columns]
            if remaining_columns:
                chunks = []
                for col in remaining_columns:
                    data_array = self.read_volume_variables(self.element, variables=[col])
                    # Filter the numpy array using the query indices
                    filtered_data_array = data_array[query_indices]
                    # Convert the filtered numpy array to a DataFrame
                    chunks.append(pd.DataFrame(filtered_data_array, columns=[col]))
                # Concatenate the query DataFrame and the remaining DataFrame
                df = pd.concat([df_query, *chunks], axis=1)
            else:
                df = df_query
        else:
            # If no query is specified, load the specified columns
            df = self.read_volume_variables(self.element, variables=columns)

        # add the index
        if with_index:
            df.set_index(self.get_index(), inplace=True, drop=True)

        return df

    def get_index(self) -> pd.MultiIndex:

        geometry: VolumeGridGeometry = self.element.geometry
        ox, oy, oz = geometry.origin

        # Make coordinates (points) along each axis, i, j, k
        i = ox + np.cumsum(geometry.tensor_u)
        i = np.insert(i, 0, ox)
        j = oy + np.cumsum(self.element.geometry.tensor_v)
        j = np.insert(j, 0, oy)
        k = oz + np.cumsum(self.element.geometry.tensor_w)
        k = np.insert(k, 0, oz)

        # convert to centroids
        x, y, z = (i[1:] + i[:-1]) / 2, (j[1:] + j[:-1]) / 2, (k[1:] + k[:-1]) / 2
        xx, yy, zz = np.meshgrid(x, y, z, indexing="ij")

        # Calculate dx, dy, dz
        dxx, dyy, dzz = np.meshgrid(geometry.tensor_u, geometry.tensor_v, geometry.tensor_w, indexing="ij")

        # TODO: consider rotation

        index = pd.MultiIndex.from_arrays([xx.ravel("F"), yy.ravel("F"), zz.ravel("F"),
                                           dxx.ravel("F"), dyy.ravel("F"), dzz.ravel("F")],
                                          names=['x', 'y', 'z', 'dx', 'dy', 'dz'])

        if len(index) != self.records_in_file:
            raise ValueError(f"The length of the index ({len(index)}) does not match the number of records"
                             f" in the VolumeElement ({self.records_in_file})")

        return index

    def read_volume_variables(self, element: str, variables: list[str]) -> pd.DataFrame:
        # Loop over the variables
        chunks: list[pd.DataFrame] = []
        for variable in variables:
            # Check if the variable exists in the VolumeElement
            if variable not in self.variables_in_file:
                raise ValueError(f"Variable '{variable}' not found in the VolumeElement: {element}")
            chunks.append(self._get_variable_by_name(variable).ravel())

        # Concatenate all chunks into a single DataFrame
        return pd.DataFrame(np.vstack(chunks), index=variables).T

    def _get_variable_by_name(self, variable_name: str):
        # get the index of the variable in order to index into elements
        variable_index = self.variables_in_file.index(variable_name)
        return self.element.data[variable_index].array.array


class ParquetFileWriter:

    def __init__(self):
        pass

    @classmethod
    def from_column_generator(cls, index: pd.Index, column_generator):

        # Path to the final output file
        output_file = "final.parquet"

        # Temp directory for storing parquet columns
        temp_dir = "temp/"

        # Ensure the temp directory exists
        os.makedirs(temp_dir, exist_ok=True)

        # Write the index to a separate Parquet file
        index_table = pa.Table.from_pandas(index.to_frame('index'))
        pq.write_table(index_table, temp_dir + "index.parquet")
        index_pf = pq.ParquetFile(temp_dir + "index.parquet")

        for i, column in enumerate(column_generator):
            # Write each column to a temporary parquet file
            table = pa.Table.from_pandas(column.to_frame())
            pq.write_table(table, temp_dir + f"column_{i}.parquet")

        # Collect paths to the temporary Parquet files
        paths = [temp_dir + file for file in os.listdir(temp_dir) if file != "index.parquet"]

        # Create a ParquetWriter for the final output file
        first_pf = pq.ParquetFile(paths[0])
        writer = pq.ParquetWriter(output_file, first_pf.schema)

        for i in range(index_pf.num_row_groups):
            # Read index chunk
            index_chunk = index_pf.read_row_group(i).to_pandas()

            # Dataframe to store chunk data
            df = pd.DataFrame(index=index_chunk['index'])

            for path in paths:
                pf = pq.ParquetFile(path)
                # Read data chunk
                data_chunk = pf.read_row_group(i).to_pandas()

                # Concatenate data chunk to the dataframe
                df = pd.concat([df, data_chunk], axis=1)

            # Write the chunk to the output file
            writer.write_table(pa.Table.from_pandas(df))

        # Close the writer and release resources
        writer.close()

        # Remove temporary files
        for file in os.listdir(temp_dir):
            os.remove(temp_dir + file)



import copy
from pathlib import Path
from typing import Optional, Literal

import pandas as pd

from elphick.geomet.base import MassComposition


class Sample(MassComposition):
    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 name: Optional[str] = None,
                 moisture_in_scope: bool = True,
                 mass_wet_var: Optional[str] = None,
                 mass_dry_var: Optional[str] = None,
                 moisture_var: Optional[str] = None,
                 component_vars: Optional[list[str]] = None,
                 composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                 components_as_symbols: bool = True,
                 ranges: Optional[dict[str, list]] = None,
                 config_file: Optional[Path] = None):
        super().__init__(data=data, name=name, moisture_in_scope=moisture_in_scope,
                         mass_wet_var=mass_wet_var, mass_dry_var=mass_dry_var,
                         moisture_var=moisture_var, component_vars=component_vars,
                         composition_units=composition_units, components_as_symbols=components_as_symbols,
                         ranges=ranges, config_file=config_file)




"""
Methods for converting volumetric data objects
REF: omfvista.volume - copied to facilitate loading selected columns/dataarrays
"""
from collections import defaultdict
from typing import TYPE_CHECKING, Optional

import numpy as np
import pyvista

if TYPE_CHECKING:
    from omf import VolumeElement

from omfvista.utilities import check_orientation


def get_volume_shape(vol):
    """Returns the shape of a gridded volume"""
    return (len(vol.tensor_u), len(vol.tensor_v), len(vol.tensor_w))


def volume_grid_geom_to_vtk(volgridgeom, origin=(0.0, 0.0, 0.0)):
    """Convert the 3D gridded volume to a :class:`pyvista.StructuredGrid`
    (or a :class:`pyvista.RectilinearGrid` when apprropriate) object contatining
    the 2D surface.

    Args:
        volgridgeom (:class:`omf.volume.VolumeGridGeometry`): the grid geometry
            to convert
    """
    volgridgeom._validate_mesh()

    ox, oy, oz = volgridgeom.origin

    # Make coordinates along each axis
    x = ox + np.cumsum(volgridgeom.tensor_u)
    x = np.insert(x, 0, ox)
    y = oy + np.cumsum(volgridgeom.tensor_v)
    y = np.insert(y, 0, oy)
    z = oz + np.cumsum(volgridgeom.tensor_w)
    z = np.insert(z, 0, oz)

    # If axis orientations are standard then use a vtkRectilinearGrid
    if check_orientation(volgridgeom.axis_u, volgridgeom.axis_v, volgridgeom.axis_w):
        return pyvista.RectilinearGrid(x + origin[0], y + origin[1], z + origin[2])

    # Otherwise use a vtkStructuredGrid
    # Build out all nodes in the mesh
    xx, yy, zz = np.meshgrid(x, y, z, indexing="ij")
    points = np.c_[xx.ravel("F"), yy.ravel("F"), zz.ravel("F")]

    # Rotate the points based on the axis orientations
    rotation_mtx = np.array([volgridgeom.axis_u, volgridgeom.axis_v, volgridgeom.axis_w])
    points = points.dot(rotation_mtx)

    output = pyvista.StructuredGrid()
    output.points = points
    output.dimensions = len(x), len(y), len(z)
    output.points += np.array(origin)
    return output


def volume_to_vtk(volelement: 'VolumeElement',
                  origin=(0.0, 0.0, 0.0),
                  columns: Optional[list[str]] = None):
    """Convert the volume element to a VTK data object.

    Args:
        volelement (:class:`omf.volume.VolumeElement`): The volume element to convert
        origin: tuple(float), optional
        columns: list[str], optional - Columns to load from the data arrays

    """
    output = volume_grid_geom_to_vtk(volelement.geometry, origin=origin)
    shp = get_volume_shape(volelement.geometry)
    # Add data to output
    if columns is None:
        for data in volelement.data:
            arr = data.array.array
            arr = np.reshape(arr, shp).flatten(order="F")
            output[data.name] = arr
    else:
        available_cols: defaultdict[str, int] = defaultdict(None, {d.name: i for i, d in enumerate(volelement.data)})
        for col in columns:
            col_index = available_cols.get(col)
            if col_index is None:
                raise ValueError(f"Column '{col}' not found in the volume element '{volelement.name}':"
                                 f" Available columns: {list(available_cols.keys())}")
            data = volelement.data[col_index]
            arr = data.array.array
            arr = np.reshape(arr, shp).flatten(order="F")
            output[data.name] = arr
    return output



import logging
from typing import Dict

import yaml


def read_yaml(file_path):
    with open(file_path, "r") as f:
        d_config: Dict = yaml.safe_load(f)
        if 'MC' != list(d_config.keys())[0]:
            msg: str = f'config file {file_path} is not a MassComposition config file - no MC key'
            logging.error(msg)
            raise KeyError(msg)
        return d_config['MC']


def read_flowsheet_yaml(file_path):
    with open(file_path, "r") as f:
        d_config: Dict = yaml.safe_load(f)
        if 'FLOWSHEET' != list(d_config.keys())[0]:
            msg: str = f'config file {file_path} is not a Flowsheet config file - no FLOWSHEET key'
            logging.error(msg)
            raise KeyError(msg)
        return d_config['FLOWSHEET']


def get_column_config(config_dict: dict, var_map: dict, config_key: str = 'range') -> dict:
    res: dict = {}
    # populate from the config
    # var_map only includes mass-composition columns, no supplementary. vars are keys, cols are values
    composition_cols = [v for k, v in var_map.items() if k not in ['mass_wet', 'mass_dry', 'moisture']]

    for k, v in config_dict['vars'].items():
        if k == 'composition':
            for col in composition_cols:
                res[col] = v[config_key]
        elif k in list(var_map.keys()) and v.get(config_key):
            res[var_map[k]] = v[config_key]
    return res



from __future__ import annotations

import functools
import uuid
from pathlib import Path
from typing import Optional, Literal, Callable, Union, Iterable, TYPE_CHECKING

import numpy as np
import pandas as pd
from pandas import IntervalIndex
from pandas.core.indexes.frozen import FrozenList

import plotly.graph_objects as go
import plotly.express as px

import elphick.geomet.flowsheet.stream
from elphick.geomet.utils.amenability import amenability_index
from elphick.geomet.utils.interp import mass_preserving_interp
from elphick.geomet.utils.interp2 import mass_preserving_interp_2d
from elphick.geomet.utils.pandas import MeanIntervalIndex, weight_average, calculate_recovery, calculate_partition, \
    cumulate, mass_to_composition

from elphick.geomet.base import MassComposition

if TYPE_CHECKING:
    from elphick.geomet.flowsheet.stream import Stream


class IntervalSample(MassComposition):
    """
    A class to represent a sample of data with an interval index.
    This exposes methods to split the sample by a partition definition.
    """

    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 name: Optional[str] = None,
                 moisture_in_scope: bool = True,
                 mass_wet_var: Optional[str] = None,
                 mass_dry_var: Optional[str] = None,
                 moisture_var: Optional[str] = None,
                 component_vars: Optional[list[str]] = None,
                 composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                 components_as_symbols: bool = True,
                 ranges: Optional[dict[str, list]] = None,
                 config_file: Optional[Path] = None):

        super().__init__(data=data, name=name, moisture_in_scope=moisture_in_scope,
                         mass_wet_var=mass_wet_var, mass_dry_var=mass_dry_var,
                         moisture_var=moisture_var, component_vars=component_vars,
                         composition_units=composition_units, components_as_symbols=components_as_symbols,
                         ranges=ranges, config_file=config_file)

        # manage the interval indexes
        if self.data is not None:
            self.data = self._create_interval_indexes(data)

    def _create_interval_indexes(self, data: pd.DataFrame) -> pd.DataFrame:
        original_indexes = data.index.names
        interval_indexes = []
        for pair in self.config['intervals']['suffixes']:
            if data.index.names != FrozenList([None]):
                suffix_candidates: dict = {n: n.split('_')[-1].lower() for n in data.index.names}
                suffixes: dict = {k: v for k, v in suffix_candidates.items() if v in pair}
                if suffixes:
                    data.reset_index(list(suffixes.keys()), inplace=True)
                    num_interval_indexes: int = int(len(suffixes.keys()) / 2)
                    for i in range(0, num_interval_indexes + 1, 2):
                        keys = list(suffixes.keys())[i: i + 2]
                        base_name: str = '_'.join(keys[0].split('_')[:-1])
                        index = IntervalIndex.from_arrays(left=data[keys[0]], right=data[keys[1]],
                                                          closed=self.config['intervals']['closed'])
                        index.name = base_name
                        # left and right names are only preserved for a single interval index.
                        # when a multiindex is used, the names are not preserved.
                        index.left.name = keys[0].split('_')[-1]
                        index.right.name = keys[1].split('_')[-1]
                        interval_indexes.append(index)

                        # drop the index columns from the dataframe columns
                        data.drop(columns=keys, inplace=True)

        if interval_indexes:
            new_indexes = {}  # Use dict to preserve order and uniqueness
            # we need to set the index to include the new interval index, but respect the order of the original.
            for i in original_indexes:
                if i.split('_')[0] not in [ii.name for ii in interval_indexes]:
                    new_indexes[i] = data.index.get_level_values(i)
                else:
                    # Find the corresponding interval index and append it to the new_indexes list
                    for ii in interval_indexes:
                        if ii.name == i.split('_')[0]:
                            new_indexes[ii.name] = ii
                            break

            if len(new_indexes) > 1:
                data.index = pd.MultiIndex.from_frame(pd.DataFrame(new_indexes.values()).T, names=new_indexes.keys())
            else:
                data.index = list(new_indexes.values())[0]

        return data

    def split_by_partition(self, partition_definition: Union[pd.Series, Callable], name_1: str = 'preferred',
                           name_2: str = 'complement'):
        """
        Split the sample into two samples based on the partition definition.

        .. math::
            K = \\frac{{m_{preferred}}}{{m_{feed}}}

        :param partition_definition: A function that takes a data frame and returns a boolean series with a
         range [0, 1]. A 1D function must have an argument that matches the dimension of the interval index.
         A 2D function must have two arguments that match the dimensions of the interval index.
        :param name_1: The name of the first sample.
        :param name_2: The name of the second sample.
        :return: A tuple of two IntervalSamples.
        """

        # Check that the partition definition has the correct number of arguments and that the names match
        sample_fraction_dims = [col for col in self.mass_data.index.names if
                                col != isinstance(self.mass_data.index.get_level_values(col), pd.IntervalIndex)]
        fraction_means: pd.DataFrame = self.mass_data.index.to_frame()[sample_fraction_dims].apply(
            lambda x: MeanIntervalIndex(x).mean, axis=0)

        # Get the function from the partial object if necessary
        if isinstance(partition_definition, Callable):
            partition_func = partition_definition.func if isinstance(partition_definition,
                                                                     functools.partial) else partition_definition
            # Check that the required argument names are present in the IntervalIndex levels
            required_args = [col for col in partition_func.__code__.co_varnames if col in sample_fraction_dims]
            pn: pd.Series = pd.Series(partition_definition(**fraction_means[required_args]), name='K',
                                      index=self._mass_data.index)
        elif isinstance(partition_definition, pd.Series):
            required_args = partition_definition.index.names
            pn: pd.Series = partition_definition
        else:
            raise TypeError(f"The partition definition must be a function or a pandas Series:"
                            f" type = {type(partition_definition)}")
        for arg, dim in zip(required_args, sample_fraction_dims):
            if arg != dim:
                raise ValueError(f"The partition definition argument name does not match the index name. "
                                 f"Expected {dim}, found {arg}")

        self.to_stream()
        self: 'Stream'

        sample_1 = self.create_congruent_object(name=name_1).to_stream()
        sample_1.mass_data = self.mass_data.copy().multiply(pn, axis=0)
        sample_1.set_nodes([self.nodes[1], uuid.uuid4()])
        sample_2 = self.create_congruent_object(name=name_2)
        sample_2.mass_data = self.mass_data.copy().multiply((1 - pn), axis=0)
        sample_2.set_nodes([self.nodes[1], uuid.uuid4()])

        return sample_1, sample_2

    def is_2d_grid(self):
        """
        Check if the sample is a 2d grid.
        :return: True if the sample has 2 levels of intervals, False otherwise.
        """
        res = False
        if self.mass_data is not None and self.mass_data.index.nlevels >= 2:
            # get the type of the index levels
            level_types = [type(level) for level in self.mass_data.index.levels]
            # get the counts of each type
            level_counts = {level_type: level_types.count(level_type) for level_type in set(level_types)}
            # check if there are 2 levels of intervals
            res = level_counts.get(pd.Interval, 0) == 2

        return res

    @property
    def is_rectilinear_grid(self):
        """If rectilinear we can plot with a simple heatmap"""
        res = False
        if self.mass_data is not None and self._mass_data.index.nlevels >= 2:
            # Get the midpoints of the intervals for X and Y
            x_midpoints = self.mass_data.index.get_level_values(0).mid
            y_midpoints = self.mass_data.index.get_level_values(1).mid

            # Get unique midpoints for X and Y
            unique_x_midpoints = set(x_midpoints)
            unique_y_midpoints = set(y_midpoints)

            # Check if the grid is full (i.e., no steps in the lines that define the grid edges)
            # todo: fix this logic - it is not correct
            if len(unique_x_midpoints) == len(x_midpoints) and len(unique_y_midpoints) == len(y_midpoints):
                res = True
        return res

    def ideal_incremental_separation(self, discard_from: Literal["lowest", "highest"] = "lowest") -> pd.DataFrame:
        """Incrementally separate a fractionated sample.

        This method sorts by the provided direction prior to incrementally removing and discarding the first fraction
         (of the remaining fractions) and recalculating the mass-composition and recovery of the portion remaining.
         This is equivalent to incrementally applying a perfect separation (partition) at every interval edge.

        This method is only applicable to a 1D object where the single dimension is a pd.Interval type.

        See also: ideal_incremental_composition, ideal_incremental_recovery.

        Args:
            discard_from: Defines the discarded direction.  discard_from = "lowest" will discard the lowest value
             first, then the next lowest, etc.

        Returns:
            A pandas DataFrame
        """
        self._check_one_dim_interval()

        sample: pd.DataFrame = self.data

        is_decreasing: bool = sample.index.is_monotonic_decreasing
        if discard_from == "lowest":
            sample.sort_index(ascending=True, inplace=True)
            new_index: pd.Index = pd.Index(sample.index.left)
        else:
            sample.sort_index(ascending=False, inplace=True)
            new_index: pd.Index = pd.Index(sample.index.right)
        new_index.name = f"{sample.index.name}_cut-point"

        aggregated_chunks: list = []
        recovery_chunks: list = []
        head: pd.Series = sample.pipe(weight_average)

        for i, indx in enumerate(sample.index):
            tmp_composition: pd.DataFrame = sample.iloc[i:, :].pipe(weight_average).to_frame().T
            aggregated_chunks.append(tmp_composition)
            recovery_chunks.append(tmp_composition.pipe(calculate_recovery, df_ref=head.to_frame().T))

        res_composition: pd.DataFrame = pd.concat(aggregated_chunks).assign(attribute="composition").set_index(
            new_index)
        res_recovery: pd.DataFrame = pd.concat(recovery_chunks).assign(attribute="recovery").set_index(
            new_index)

        if is_decreasing:
            res_composition.sort_index(ascending=False, inplace=True)
            res_recovery.sort_index(ascending=False, inplace=True)

        res: pd.DataFrame = pd.concat([res_composition, res_recovery]).reset_index().set_index(
            [new_index.name, 'attribute'])

        return res

    def _check_one_dim_interval(self):
        if len(self.mass_data.index.names) != 1:
            raise NotImplementedError(f"This object is {self.mass_data.index.ndim} dimensional. "
                                      f"Only 1D interval objects are valid")
        index_var: str = self.mass_data.index.name
        if not isinstance(self.mass_data.index, pd.IntervalIndex):
            raise NotImplementedError(f"The {index_var} of this object is not a pd.Interval. "
                                      f" Only 1D interval objects are valid")

    def _check_two_dim_interval(self):
        if len(self.mass_data.index.names) != 2:
            raise NotImplementedError(f"This object is {self.mass_data.index.ndim} dimensional. "
                                      f"Only 2D interval objects are valid")
        for indx in self.mass_data.index.levels:
            if not isinstance(indx, pd.IntervalIndex):
                raise NotImplementedError(f"The {indx.name} of this object is not a pd.Interval. "
                                          f" Only 1D interval objects are valid")

    def ideal_incremental_composition(self, discard_from: Literal["lowest", "highest"] = "lowest") -> pd.DataFrame:
        """Incrementally separate a fractionated sample.

        This method sorts by the provided direction prior to incrementally removing and discarding the first fraction
         (of the remaining fractions) and recalculating the mass-composition of the portion remaining.
         This is equivalent to incrementally applying a perfect separation (partition) at every interval edge.

        This method is only applicable to a 1D object where the single dimension is a pd.Interval type.

        See also: ideal_incremental_separation, ideal_incremental_recovery.

        Args:
            discard_from: Defines the discarded direction.  discard_from = "lowest" will discard the lowest value
             first, then the next lowest, etc.

        Returns:
            A pandas DataFrame
        """
        df: pd.DataFrame = self.ideal_incremental_separation(discard_from=discard_from).query(
            'attribute=="composition"').droplevel('attribute')
        return df

    def ideal_incremental_recovery(self, discard_from: Literal["lowest", "highest"] = "lowest",
                                   apply_closure: bool = True) -> pd.DataFrame:
        """Incrementally separate a fractionated sample.

        This method sorts by the provided direction prior to incrementally removing and discarding the first fraction
         (of the remaining fractions) and recalculating the recovery of the portion remaining.
         This is equivalent to incrementally applying a perfect separation (partition) at every interval edge.

        This method is only applicable to a 1D object where the single dimension is a pd.Interval type.

        See also: ideal_incremental_separation, ideal_incremental_composition.

        Args:
            discard_from: Defines the discarded direction.  discard_from = "lowest" will discard the lowest value
             first, then the next lowest, etc.
            apply_closure: If True, Add the missing record (zero recovery) that closes the recovery envelope.

        Returns:
            A pandas DataFrame
        """
        columns_to_drop: list[str] = ['mass_wet', 'H2O'] if self.moisture_in_scope else []
        df: pd.DataFrame = self.ideal_incremental_separation(discard_from=discard_from).query(
            'attribute=="recovery"').droplevel('attribute').rename(columns={'mass_dry': 'mass'}).drop(
            columns=columns_to_drop)
        if apply_closure:
            # add zero recovery record to close the envelope.
            indx = np.inf if df.index.min() == 0.0 else 0.0
            indx_name: str = df.index.name
            df = pd.concat([df, pd.Series(0, index=df.columns, name=indx).to_frame().T]).sort_index(ascending=True)
            df.index.name = indx_name
        return df

    def plot_heatmap(self, components: list[str], **kwargs):
        """
        Plot the sample as a heatmap.
        :param components: The list of components to plot.
        :param kwargs: Additional keyword arguments to pass to the plot method.
        :return: The axis with the plot.
        """
        # if not self.is_rectilinear_grid:
        #     raise ValueError('The sample is not a rectilinear grid.')

        # convert IntervalIndex to nominal values df.index = df.index.map(lambda x: x.mid)

        x_label = self.mass_data.index.names[0]
        y_label = self.mass_data.index.names[1]
        z_label = self.mass_data.columns[0]

        # create a pivot table for the heatmap
        pivot_df = self.mass_data[components].copy().unstack()

        # Get the midpoints of the intervals for X and Y
        x_midpoints = [interval.mid for interval in self.mass_data.index.get_level_values(x_label)]
        y_midpoints = [interval.mid for interval in self.mass_data.index.get_level_values(y_label)]

        # Get interval edges for x and y axes
        x_edges = self._get_unique_edges(self.mass_data.index.get_level_values(x_label))
        y_edges = self._get_unique_edges(self.mass_data.index.get_level_values(y_label))

        # Create hover text
        hover_text = [[f"{x_label}: {x_mid}, {y_label}: {y_mid}, {z_label}: {z_val}"
                       for x_mid, z_val in zip(x_midpoints, z_values)]
                      for y_mid, z_values in zip(y_midpoints, pivot_df.values)]

        # plot the heatmap
        fig = go.Figure(data=go.Heatmap(
            z=pivot_df.values,
            x=x_edges,
            y=y_edges,
            text=hover_text,
            hoverinfo='text'))

        # update the layout to use logarithmic x-axis
        if x_label == 'size':
            fig.update_layout(xaxis_type="log")
        elif y_label == 'size':
            fig.update_layout(yaxis_type="log")

        # set the title and x and y labels dynamically
        fig.update_layout(title=f'{self.name} Heatmap',
                          xaxis_title=self.mass_data.index.names[0],
                          yaxis_title=self.mass_data.index.names[1])

        return fig

    def plot_intervals(self,
                       variables: list[str],
                       cumulative: bool = True,
                       direction: str = 'descending',
                       show_edges: bool = True,
                       min_x: Optional[float] = None) -> go.Figure:
        """Plot "The Grade-Tonnage" curve.

        Mass and grade by bins for a cut-off variable.

        Args:
            variables: List of variables to include in the plot
            cumulative: If True, the results are cumulative weight averaged.
            direction: 'ascending'|'descending', if cumulative is True, the direction of accumulation
            show_edges: If True, show the edges on the plot.  Applicable to cumulative plots only.
            min_x: Optional minimum value for the x-axis, useful to set reasonable visual range with a log
            scaled x-axis when plotting size data
        """

        res: pd.DataFrame = self.data[variables]

        plot_kwargs: dict = dict(line_shape='vh')
        if cumulative:
            res = self.mass_data.pipe(cumulate, direction=direction).pipe(mass_to_composition)
            plot_kwargs = dict(line_shape='spline')

        interval_data: pd.DataFrame = res

        # Get the first IntervalIndex - TODO: specify or check...
        interval_index: Optional[pd.IntervalIndex] = None
        for level in range(interval_data.index.nlevels):
            if isinstance(interval_data.index.get_level_values(level), pd.IntervalIndex):
                interval_index = interval_data.index.get_level_values(level)
                break
        if interval_index is None:
            raise ValueError("No IntervalIndex found in the index levels")
        left_name: str = interval_index.left.name if interval_index.left.name else 'left'
        right_name: str = interval_index.right.name if interval_index.right.name else 'right'
        left: pd.Series = pd.Series(interval_index.left, name=left_name, index=interval_index)
        right: pd.Series = pd.Series(interval_index.right, name=right_name, index=interval_index)
        df_intervals = pd.concat([left, right, interval_data], axis='columns')
        x_var: str = interval_data.index.name
        if not cumulative:
            # append on the largest fraction right edge for display purposes
            is_ascending: bool = interval_index.is_monotonic_increasing
            df_end: pd.DataFrame = df_intervals.loc[df_intervals.index.max(), :].to_frame().T
            df_end[left_name] = df_end[right_name]
            df_end[right_name] = np.inf
            df = pd.concat([df_end.reset_index(drop=True), df_intervals], axis='index')
            df[interval_data.index.name] = df[left_name]
            df = df.sort_values(by=interval_data.index.name, ascending=is_ascending)
        else:
            if direction == 'ascending':
                x_var = right_name
            elif direction == 'descending':
                x_var = left_name
            df = df_intervals

        if res.index.name.lower() == 'size':
            if not min_x:
                min_x = interval_data.index.min().right / 2.0
            # set zero to the minimum x value (for display only) to enable the tooltips on that point.
            df.loc[df[x_var] == df[x_var].min(), x_var] = min_x
            hover_data = {'component': True,  # add other column, default formatting
                          x_var: ':.3f',  # add other column, customized formatting
                          'value': ':.2f'
                          }
            plot_kwargs = {**plot_kwargs,
                           **dict(log_x=True,
                                  range_x=[min_x, interval_data.index.max().right],
                                  hover_data=hover_data)}

        df = df[[x_var] + variables].melt(id_vars=[x_var], var_name='component')

        if cumulative and show_edges:
            plot_kwargs['markers'] = True

        fig = px.line(df, x=x_var, y='value', facet_row='component', **plot_kwargs)
        fig.for_each_annotation(lambda a: a.update(text=a.text.replace("component=", "")))
        fig.update_yaxes(matches=None)
        fig.update_layout(title=self.name)

        return fig

    @staticmethod
    def _get_unique_edges(interval_index):
        # Get the left and right edges of the intervals
        left_edges = interval_index.left.tolist()
        right_edges = interval_index.right.tolist()

        # Concatenate the two lists
        all_edges = left_edges + right_edges

        # Get the unique edges
        unique_edges = np.unique(all_edges)

        return unique_edges

    def plot_grade_recovery(self, target_analyte,
                            discard_from: Literal["lowest", "highest"] = "lowest",
                            title: Optional[str] = None,
                            ) -> go.Figure:
        """The grade-recovery plot.

        The grade recovery curve is generated by assuming an ideal separation (for the chosen property, or dimension)
        at each fractional interval.  It defines the theoretical maximum performance, which can only be improved if
        liberation is improved by comminution.

        This method is only applicable to a 1D object where the single dimension is a pd.Interval type.

        Args:
            target_analyte: The analyte of value.
            discard_from: Defines the discarded direction.  discard_from = "lowest" will discard the lowest value
             first, then the next lowest, etc.
            title: Optional plot title

        Returns:
            A plotly.GraphObjects figure
        """
        title = title if title is not None else 'Ideal Grade - Recovery'
        cols_to_drop: list[str] = ['mass_wet', 'mass_dry', 'H2O'] if self.moisture_in_scope else ['mass_dry']

        df: pd.DataFrame = self.ideal_incremental_separation(discard_from=discard_from)
        df_recovery: pd.DataFrame = df.loc[(slice(None), 'recovery'), [target_analyte, 'mass_dry']].droplevel(
            'attribute').rename(
            columns={'mass_dry': 'Yield', target_analyte: f"{target_analyte}_recovery"})
        df_composition: pd.DataFrame = df.loc[(slice(None), 'composition'), :].droplevel('attribute').drop(
            columns=cols_to_drop)

        df_plot: pd.DataFrame = pd.concat([df_recovery, df_composition], axis=1).reset_index()
        fig = px.line(df_plot, x=target_analyte,
                      y=f"{target_analyte}_recovery",
                      hover_data=df_plot.columns,
                      title=title)
        # fig.update_layout(xaxis_title=f"Grade of {target_analyte}", yaxis_title=f"Recovery of {target_analyte}",
        #                   title=title)

        return fig

    def plot_amenability(self, target_analyte: str,
                         discard_from: Literal["lowest", "highest"] = "lowest",
                         gangue_analytes: Optional[str] = None,
                         title: Optional[str] = None,
                         ) -> go.Figure:
        """The yield-recovery plot.

        The yield recovery curve provides an understanding of the amenability of a sample.

        This method is only applicable to a 1D object where the single dimension is a pd.Interval type.

        Args:
            target_analyte: The analyte of value.
            discard_from: Defines the discarded direction.  discard_from = "lowest" will discard the lowest value
             first, then the next lowest, etc.
            gangue_analytes: The analytes to be rejected
            title: Optional plot title

        Returns:
            A plotly.GraphObjects figure
        """
        title = title if title is not None else 'Amenability Plot'
        df: pd.DataFrame = self.ideal_incremental_recovery(discard_from=discard_from)
        amenability_indices: pd.Series = amenability_index(df, col_target=target_analyte, col_mass_recovery='mass')

        analytes = [col for col in df.columns if col != "mass"] if gangue_analytes is None else [
            target_analyte + gangue_analytes]

        mass_rec: pd.DataFrame = df["mass"]
        df = df[analytes]

        fig = go.Figure()
        for analyte in analytes:
            fig.add_trace(
                go.Scatter(x=mass_rec, y=df[analyte], mode="lines",
                           name=f"{analyte} ({round(amenability_indices[analyte], 2)})",
                           customdata=df.index.values,
                           hovertemplate='<b>Recovery: %{y:.3f}</b><br>Cut-point: %{customdata:.3f} '))
        fig.add_trace(go.Scatter(x=[0, 1], y=[0, 1], mode="lines", name='y=x',
                                 line=dict(shape='linear', color='gray', dash='dash'),
                                 ))
        fig.update_layout(xaxis_title='Yield (Mass Recovery)', yaxis_title='Recovery', title=title,
                          hovermode='x')
        return fig

    def calculate_partition(self, preferred: 'MassComposition') -> pd.DataFrame:
        """Calculate the partition number (K) [0, 1] of the preferred stream relative to self

        .. math::
            K = \\frac{{m_{preferred}}}{{m_{feed}}}

        """
        self._check_one_dim_interval()
        return calculate_partition(df_feed=self.data, df_preferred=preferred.data,
                                   col_mass_dry='mass_dry')

    def resample_1d(self, interval_edges: Union[Iterable, int],
                    precision: Optional[int] = None,
                    include_original_edges: bool = False) -> 'IntervalSample':
        """Resample a 1D fractional dim/index

        Args:
            interval_edges: The values of the new grid (interval edges).  If an int, will up-sample by that factor, for
             example the value of 10 will automatically define edges that create 10 x the resolution (up-sampled).
            precision: Optional integer for the number of decimal places to round the grid values to.
            include_original_edges: If True include the original edges in the grid.

        Returns:
            A new IntervalSample object interpolated onto the new grid
        """

        # TODO: add support for supplementary variables

        # test the index contains a single interval index
        self._check_one_dim_interval()

        df_upsampled: pd.DataFrame = mass_preserving_interp(self.mass_data,
                                                            interval_edges=interval_edges, precision=precision,
                                                            include_original_edges=include_original_edges,
                                                            mass_wet=None, mass_dry=self.mass_dry_var,
                                                            interval_data_as_mass=True)

        obj: IntervalSample = IntervalSample(df_upsampled, name=self.name, moisture_in_scope=False,
                                             mass_dry_var=self.mass_dry_var)
        obj.status.ranges = self.status.ranges
        return obj

    def resample_2d(self, interval_edges: dict[str, Iterable],
                    precision: Optional[int] = None) -> 'IntervalSample':
        """Resample a 2D fractional dim/index

        Args:
            interval_edges: A dict keyed by index name containing the grid the data is resampled to.
            precision: Optional integer for the number of decimal places to round the grid values to.

        Returns:
            A new IntervalSample object interpolated onto the new grid
        """

        # TODO: add support for supplementary variables

        # test the index contains a single interval index
        self._check_two_dim_interval()

        df_upsampled_specific_mass: pd.DataFrame = mass_preserving_interp_2d(self._specific_mass(),
                                                                             interval_edges=interval_edges,
                                                                             precision=precision,
                                                                             mass_dry=self.mass_dry_var)

        # convert from specific mass to mass
        df_upsampled = df_upsampled_specific_mass.mul(self.mass_data[self.mass_dry_var].sum(), axis=0)
        df_upsampled[self.composition_columns] = df_upsampled[self.composition_columns].div(
            df_upsampled[self.mass_dry_var], axis=0).mul(self.composition_factor, axis=0)

        obj: IntervalSample = IntervalSample(df_upsampled, name=self.name, moisture_in_scope=False,
                                             mass_dry_var=self.mass_dry_var)
        if hasattr(obj, 'nodes'):
            obj.nodes = self.nodes
        obj.status.ranges = self.status.ranges
        return obj

    def _specific_mass(self) -> Optional[pd.DataFrame]:
        """Calculate the specific mass of the sample

        Specific mass is the mass of the sample fractions divided by the mass of all fractions.
        The sum of the specific mass (for mass_dry) is 1.0 by definition.
        """
        res = None
        if self.data is not None:
            res = self.mass_data.div(self.mass_data[self.mass_dry_var].sum(), axis=0)
            if self.moisture_in_scope:
                res.drop(columns=[self.mass_wet_var], inplace=True)
        return res



from typing import Optional, Iterable, Union

import numpy as np
import pandas as pd
from scipy.interpolate import pchip_interpolate, PchipInterpolator

from elphick.geomet.utils.pandas import composition_to_mass, mass_to_composition, weight_average


def mass_preserving_interp_2d(specific_mass_intervals: pd.DataFrame,
                              interval_edges: Optional[dict[str, Iterable[float]]] = None,
                              precision: Optional[int] = None,
                              mass_dry: str = 'mass_dry') -> pd.DataFrame:
    """Interpolate 2D interval data with zero mass loss using pchip

    This function applies mass-preserving up-sampling to 2D interval data.  The function will:
    For example if the first dimension is size and the second is density:

    1. Convert to specific mass
    2. Accumulate along both dimensions
    3. For every first (size) dimension value
        a. fit a monotonic spline for the values of the second (density) dimension
        b. interpolate the second (density) dimension to the supplied (density) grid
    4. For every value on the second (density) upsampled (supplied) grid:
        a. fit a monotonic spline to the values of the first (size) dimension
        b. interpolate the first (size) dimension to the supplied (size) grid
    5. Decumulate along both dimensions
    6. Convert to composition

    The above steps are repeated for every component, starting with dry mass and ending with the last component.


    Args:
        specific_mass_intervals: Dataframe with two pd.IntervalIndexes, representing specific mass intervals.
        interval_edges: Dict of the values of the new grid (interval edges) for each dimension, keyed by index
         name (dimension).  If None, the grid will be a rectilinear grid using unique values from the index.
        precision: Number of decimal places to round the index (edge) values.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.

    Returns:

    """

    if precision is None:
        precision = 6

    dim_names = specific_mass_intervals.index.names
    original_dim1_vals: np.ndarray = np.unique(specific_mass_intervals.index.get_level_values(dim_names[0]).right)
    original_dim2_vals: np.ndarray = np.unique(specific_mass_intervals.index.get_level_values(dim_names[1]).right)

    specific_mass_intervals.sort_index(ascending=[True, True], inplace=True)

    if interval_edges is None:
        interval_edges: dict[str, list[float]] = {dim: specific_mass_intervals.index.get_level_values(dim).unique()
                                                  for dim in dim_names}

    # check that the supplied interval edges are bounded on the left by the sampled minimum
    for dim in dim_names:
        if np.min(interval_edges[dim]) < specific_mass_intervals.index.get_level_values(dim).left.min():
            raise ValueError(f"The supplied {dim} grid contains values lower than the minimum in the sample.")


    # Convert to a numpy 3d array
    # [i, j, k] where i is the dim1, j is the dim2, and k is the component

    # We will accumulate along both dimensions, but notably we cannot assume that the 2d data sits on a
    # rectilinear grid.

    # TODO: test this code block in the case of non-rectilinear input data.
    mass_data: np.ndarray = specific_mass_intervals.to_numpy().reshape(
        specific_mass_intervals.index.get_level_values(dim_names[0]).unique().size,
        specific_mass_intervals.index.get_level_values(dim_names[1]).unique().size, -1)

    mass_data_cum: np.ndarray = np.cumsum(np.cumsum(mass_data, axis=0), axis=1)

    # create a 3D array to store the upsampled dim2 data (for each original dim1 value)
    upsampled_dim2: np.ndarray = np.zeros(
        (len(original_dim1_vals), len(interval_edges[dim_names[1]]), mass_data_cum.shape[2]))

    # create a 3D array to store the full upsampled (dim1 and dim2) data
    mass_data_upsampled = np.zeros(
        (len(interval_edges[dim_names[0]]), len(interval_edges[dim_names[1]]), mass_data_cum.shape[2]))

    for k in range(mass_data_cum.shape[2]):  # for every component (c)

        for i in range(mass_data_cum.shape[0]):  # for every dim1 (i), upsample dim2 (j)
            x = original_dim2_vals  # dim2
            y = mass_data_cum[i, :, k]  # component

            dim2_spline = PchipInterpolator(x, y, extrapolate=True)
            # now we can upsample the spline to a finer grid
            y_new = dim2_spline(interval_edges[dim_names[1]])  # upsampled dim2
            # Store the upsampled data
            upsampled_dim2[i, :, k] = y_new

        # for every upsampled dim2 value (j), upsample dim1 (i)
        for j in range(len(interval_edges[dim_names[1]])):
            x = original_dim1_vals  # dim1
            y = upsampled_dim2[:, j, k]  # component

            dim1_spline = PchipInterpolator(x, y, extrapolate=True)
            # now we can upsample the spline to a finer grid
            y_new = dim1_spline(interval_edges[dim_names[0]])  # upsampled size
            # Store the upsampled data
            mass_data_upsampled[:, j, k] = y_new

    print(mass_data_upsampled.shape)
    # Now we can decumulate along both dimensions
    mass_data_decum: np.ndarray = np.diff(
        np.diff(mass_data_upsampled, axis=0,
                prepend=np.zeros((1, mass_data_upsampled.shape[1], mass_data_upsampled.shape[2]))),
        axis=1, prepend=np.zeros((mass_data_upsampled.shape[0], 1, mass_data_upsampled.shape[2])))

    # Reshape to a 2D array
    mass_data_decum = mass_data_decum.reshape(mass_data_upsampled.shape[0] * mass_data_upsampled.shape[1],
                                              mass_data_upsampled.shape[2])

    # Create the multiindex of IntervalIndex for the upsampled data, using the supplied interval edges

    # Generate the MultiIndex
    min_left_edges = np.array([i.left for i in specific_mass_intervals.index[0]])
    dim1_indexes = pd.IntervalIndex.from_breaks(
        np.round(np.hstack((min_left_edges[0], interval_edges[dim_names[0]])), precision),
        closed='left')
    dim2_indexes = pd.IntervalIndex.from_breaks(
        np.round(np.hstack((min_left_edges[1], interval_edges[dim_names[1]])), precision),
        closed='left')

    index = pd.MultiIndex.from_product([dim1_indexes, dim2_indexes], names=dim_names)

    # TODO: account for uniform density for some sizes, like cyclosize.
    # This can be by 3 splines, one for mid, and the two boundaries, by extrapolation.

    return pd.DataFrame(mass_data_decum, index=index, columns=specific_mass_intervals.columns)



import pandas as pd


def sample_data(include_wet_mass: bool = True, include_dry_mass: bool = True,
                include_moisture: bool = False, include_chem_vars: bool = True) -> pd.DataFrame:
    """Creates synthetic data for testing

    Args:
        include_wet_mass: If True, wet mass is included.
        include_dry_mass: If True, dry mass is included.
        include_moisture: If True, moisture (H2O) is included.
        include_chem_vars: If True, chemical variables are included.

    Returns:

    """

    # mass_wet: pd.Series = pd.Series([100, 90, 110], name='wet_mass')
    # mass_dry: pd.Series = pd.Series([90, 80, 100], name='dry_mass')
    mass_wet: pd.Series = pd.Series([100., 90., 110.], name='wet_mass')
    mass_dry: pd.Series = pd.Series([90., 80., 90.], name='mass_dry')
    chem: pd.DataFrame = pd.DataFrame.from_dict({'FE': [57., 59., 61.],
                                                 'SIO2': [5.2, 3.1, 2.2],
                                                 'al2o3': [3.0, 1.7, 0.9],
                                                 'LOI': [5.0, 4.0, 3.0]})
    attrs: pd.Series = pd.Series(['grp_1', 'grp_1', 'grp_2'], name='group')

    if include_wet_mass and not include_dry_mass:
        mass = pd.DataFrame(mass_wet)
    elif not include_wet_mass and include_dry_mass:
        mass = pd.DataFrame(mass_dry)
    elif include_wet_mass and include_dry_mass:
        mass = pd.concat([mass_wet, mass_dry], axis='columns')
    else:
        raise AssertionError('Arguments provided result in no mass column')

    if include_moisture is True:
        moisture: pd.DataFrame = (mass_wet - mass_dry) / mass_wet * 100
        moisture.name = 'H2O'
        res: pd.DataFrame = pd.concat([mass, moisture, chem, attrs], axis='columns')
    else:
        res: pd.DataFrame = pd.concat([mass, chem, attrs], axis='columns')

    if include_chem_vars is False:
        res = res.drop(columns=chem.columns)

    res.index.name = 'index'

    return res



from copy import copy
from enum import Enum
from functools import reduce
from pathlib import Path
from typing import Optional, TypeVar, TYPE_CHECKING

import numpy as np
import pandas as pd
import yaml

from elphick.geomet.base import MC
from elphick.geomet.flowsheet.stream import Stream
from elphick.geomet.utils.pandas import MeanIntervalIndex
from elphick.geomet.utils.partition import load_partition_function

# generic type variable, used for type hinting that play nicely with subclasses
OP = TypeVar('OP', bound='Operation')

if TYPE_CHECKING:
    from elphick.geomet import IntervalSample


class NodeType(Enum):
    SOURCE = 'input'
    SINK = 'output'
    BALANCE = 'degree 2+'


class Operation:
    def __init__(self, name, config_file: Optional[Path] = None):
        self.name = name
        self._inputs = []
        self._outputs = []
        self._is_balanced: Optional[bool] = None
        self._unbalanced_records: Optional[pd.DataFrame] = None

        # Load configuration
        if config_file is None:
            config_file = Path(__file__).parent / '.././config/mc_config.yml'

        with open(config_file, 'r') as file:
            config = yaml.safe_load(file)
        self.tol_abs = config['flowsheet']['operation'].get('tol_abs', 1e-6)
        self.tol_rel = config['flowsheet']['operation'].get('tol_rel', 1e-6)

    @property
    def has_empty_input(self) -> bool:
        return None in self.inputs

    @property
    def has_empty_output(self) -> bool:
        return None in self.outputs

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, value: list[MC]):
        self._inputs = value
        self.check_balance()

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value: list[MC]):
        self._outputs = value
        self.check_balance()

    @property
    def node_type(self) -> Optional[NodeType]:
        if self.inputs and not self.outputs:
            res = NodeType.SINK
        elif self.outputs and not self.inputs:
            res = NodeType.SOURCE
        elif self.inputs and self.outputs:
            res = NodeType.BALANCE
        else:
            res = None
        return res

    def get_input_mass(self) -> pd.DataFrame:
        inputs = [i for i in self.inputs if i is not None]

        if not inputs:
            try:
                return self._create_zero_mass()
            except:
                return pd.DataFrame()
        elif len(inputs) == 1:
            return inputs[0].mass_data
        else:
            return reduce(lambda a, b: a.add(b, fill_value=0), [stream.mass_data for stream in inputs])

    def get_output_mass(self) -> pd.DataFrame:
        outputs = [o for o in self.outputs if o is not None]

        if not outputs:
            try:
                return self._create_zero_mass()
            except:
                return pd.DataFrame()
        elif len(outputs) == 1:
            return outputs[0].mass_data
        else:
            return reduce(lambda a, b: a.add(b, fill_value=0), [output.mass_data for output in outputs])

    def check_balance(self):
        """Checks if the mass and chemistry of the input and output are balanced"""
        if not self.inputs or not self.outputs:
            return None

        input_mass, output_mass = self.get_input_mass(), self.get_output_mass()
        mask = ~np.isclose(input_mass, output_mass, atol=self.tol_abs, rtol=self.tol_rel)
        self._unbalanced_records = (input_mass - output_mass).loc[mask.any(axis=1)]
        self._is_balanced = self._unbalanced_records.shape[0] == 0

    @property
    def is_balanced(self) -> Optional[bool]:
        return self._is_balanced

    @property
    def unbalanced_records(self) -> Optional[pd.DataFrame]:
        return self._unbalanced_records

    def solve(self) -> Optional[MC]:
        """Solves the operation

        Missing data is represented by None in the input and output streams.
        Solve will replace None with an object that balances the mass and chemistry of the input and output streams.
        Returns
        The back-calculated mc object
        """

        # Check the number of missing inputs and outputs
        missing_count: int = self.inputs.count(None) + self.outputs.count(None)
        if missing_count > 1:
            raise ValueError("The operation cannot be solved - too many degrees of freedom")
        mc = None
        if missing_count == 0 and self.is_balanced:
            return mc
        else:
            if None in self.inputs:
                ref_object = self.outputs[0]
                # Find the index of None in inputs
                none_index = self.inputs.index(None)

                # Calculate the None object
                new_input_mass: pd.DataFrame = self.get_output_mass() - self.get_input_mass()
                # Create a new object from the mass dataframe
                mc = type(ref_object).from_mass_dataframe(new_input_mass, mass_wet=ref_object.mass_wet_var,
                                                          mass_dry=ref_object.mass_dry_var,
                                                          moisture_column_name=ref_object.moisture_column,
                                                          component_columns=ref_object.composition_columns,
                                                          composition_units=ref_object.composition_units,
                                                          moisture_in_scope=ref_object.moisture_in_scope)
                # Replace None with the new input
                self.inputs[none_index] = mc

            elif None in self.outputs:
                ref_object = self.inputs[0]
                # Find the index of None in outputs
                none_index = self.outputs.index(None)

                # Calculate the None object
                if len(self.outputs) == 1 and len(self.inputs) == 1:
                    # passthrough, no need to calculate.  Shallow copy to minimise memory.
                    mc = copy(self.inputs[0])
                    mc.name = None
                else:
                    new_output_mass: pd.DataFrame = self.get_input_mass() - self.get_output_mass()
                    # Create a new object from the mass dataframe
                    mc = type(ref_object).from_mass_dataframe(new_output_mass, mass_wet=ref_object.mass_wet_var,
                                                              mass_dry=ref_object.mass_dry_var,
                                                              moisture_column_name=ref_object.moisture_column,
                                                              component_columns=ref_object.composition_columns,
                                                              composition_units=ref_object.composition_units,
                                                              moisture_in_scope=ref_object.moisture_in_scope)

                # Replace None with the new output
                self.outputs[none_index] = mc

            # update the balance related attributes
            self.check_balance()
            return mc

    def _create_zero_mass(self) -> pd.DataFrame:
        """Creates a zero mass dataframe with the same columns and index as the mass data"""
        # get the firstan object with the mass data
        obj = self._get_object()
        return pd.DataFrame(data=0, columns=obj.mass_data.columns, index=obj.mass_data.index)

    def _get_object(self, name: Optional[str] = None) -> MC:
        """Returns an object from inputs or outputs"""
        candidates = [mc for mc in self.outputs + self.inputs if mc is not None]
        if len(candidates) == 0:
            raise ValueError("No object found")
        if name:
            for obj in candidates:
                if obj is not None and obj.name == name:
                    return obj
            raise ValueError(f"No object found with name {name}")
        else:
            return candidates[0]

    @classmethod
    def from_dict(cls, config: dict) -> 'Operation':
        name = config.get('name')

        return cls(name=name)


class Input(Operation):
    def __init__(self, name):
        super().__init__(name)


class Output(Operation):
    def __init__(self, name):
        super().__init__(name)


class Passthrough(Operation):
    def __init__(self, name):
        super().__init__(name)


class PartitionOperation(Operation):
    """An operation that partitions the input stream into multiple output streams based on a partition function

    The partition input is the mean of the fractions or the geomean if the fractions are in the size dimension
    The partition function is typically a partial function so that the partition is defined for all arguments
    other than the input mean fraction values in one or two dimensions.  The argument names must match the
    index names in the IntervalSample.

    """

    def __init__(self, name, partition=None):
        super().__init__(name)
        self.partition = partition
        self.partition_function = None
        if self.partition and 'module' in self.partition and 'function' in self.partition:
            self.partition_function = load_partition_function(self.partition['module'], self.partition['function'])

    def solve(self) -> [MC, MC]:
        if self.partition_function:
            self.apply_partition()
            # update the balance related attributes
            self.check_balance()
        return self.outputs

    def apply_partition(self):
        if len(self.inputs) != 1:
            raise ValueError("PartitionOperation must have exactly one input")
        for input_sample in self.inputs:
            input_sample: 'IntervalSample'
            if input_sample is not None:
                output, complement = input_sample.split_by_partition(self.partition_function)
                self.outputs = [output, complement]

    @classmethod
    def from_dict(cls, config: dict) -> 'PartitionOperation':
        name = config.get('name')
        partition = config.get('partition')
        return cls(name=name, partition=partition)



import uuid

from elphick.geomet.base import MassComposition


class Stream(MassComposition):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nodes = [uuid.uuid4(), uuid.uuid4()]

    def set_parent_node(self, parent: 'Stream') -> 'Stream':
        self.nodes = [parent.nodes[1], self.nodes[1]]
        return self

    def set_child_node(self, child: 'Stream') -> 'Stream':
        self.nodes = [self.nodes[0], child.nodes[0]]
        return self

    def set_nodes(self, nodes: list) -> 'Stream':
        if len(nodes) != 2:
            raise ValueError('Nodes must be a list of length 2')
        if nodes[0] == nodes[1]:
            raise ValueError('Nodes must be different')
        self.nodes = nodes
        return self

    # @classmethod
    # def from_mass_composition(cls, obj: MassComposition) -> 'Stream':
    #     filtered_kwargs = filter_kwargs(obj, **obj.__dict__)
    #     filtered_kwargs['data'] = obj.data
    #     stream = cls(**filtered_kwargs)
    #     stream.__class__ = type(obj.__class__.__name__, (obj.__class__, cls), {})
    #     return stream

    @classmethod
    def from_dict(cls, config: dict) -> 'Stream':
        name = config.get('name')
        node_in = config.get('node_in')
        node_out = config.get('node_out')
        return cls(name=name).set_nodes([node_in, node_out])


from __future__ import annotations
from importlib import metadata

from .sample import Sample
from .interval_sample import IntervalSample

try:
    __version__ = metadata.version('geometallurgy')
except metadata.PackageNotFoundError:
    # Package is not installed
    pass
